/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2016 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Application
    IBSBM

Description
    Transient solver for general 2D incompressible suspensions flow of spherical particles over rigid obstacles.

For more details, see:

T. DBOUK, A suspension balance direct-forcing immersed boundary model for wet granular flows over obstacles, Journal of Non-Newtonian Fluid Mechanics 230 (2016) 68–79.

 
------------------------------------------------
		A Solver by:
------------------------------------------------
                  Dr. Talib DBOUK 
		(+33) 06 50 85 15 90
		(+33) 03 27 71 23 90
  		www.talibdbouk.com
          Emails: talib.dbouk@mines-douai.fr
	          email@talibdbouk.com
	
------------------------------------------------
		Professional Address:
------------------------------------------------
	   Ecole des Mines de Douai
	   941 rue Charles Bourseul
	        CS10838
             59508 DOUAI, FRANCE
------------------------------------------------

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "pisoControl.H"

#include <iostream>
#include <fstream>


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{

#   include "setRootCase.H"

#   include "createTime.H"
#   include "createMesh.H"

    pisoControl piso(mesh);

#   include "createFields.H"
#   include "initContinuityErrs.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int nbc=0;

   

//suspension phase - fill with smooth function above initial BC values c=0.56 
dimensionedScalar xdim("xdim", dimensionSet(0,0,0, 0, 0, 0, 0), 0.0);  // mesh x position
dimensionedScalar ydim("ydim", dimensionSet(0,0,0, 0, 0, 0, 0), 0.0);  // mesh y position
dimensionedScalar zdim("zdim", dimensionSet(0,0,0, 0, 0, 0, 0), 0.0);  // mesh z position
dimensionedScalar xpdim("xpdim", dimensionSet(0,0,0, 0, 0, 0, 0), 0.0); // suspension x concentration center
dimensionedScalar conc_value("conc_value", dimensionSet(0,0,0, 0, 0, 0, 0), 0.0); // to be smoothed


    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

    Info<< "\nStarting time loop\n" << endl;
    	int count = 0;
	std::ofstream out("FSI.txt");
        out << "#time " << "FSI_x " << "FSI_y " << "FSI_z " << "Us_x " << "Us_y " << "Us_z " << std::endl;

	// boundaries

	label b1 = mesh.boundaryMesh().findPatchID("inlet");  label b2 = mesh.boundaryMesh().findPatchID("outlet");
	label b3 = mesh.boundaryMesh().findPatchID("lowerWall");  label b4 = mesh.boundaryMesh().findPatchID("upperWall");


	int COUNT = 3 ;

	// (xc,yc,zc)=Center and Radius=radius of an immersed sphere. Are to get from user in transportProperties file


	// Interaction Force interacting between each (n and n+1) time steps (velocity-verlet algorithm)
	double FS_X = 0.; double FS_Y = 0.; double FS_Z = 0.;

	// Solid velocity interacting between each (n and n+1) time steps (velocity-verlet algorithm)
	double US_X = 0.; double US_Y = 0.; double US_Z = 0.;


	// Initial gravity to fill in (g_X, g_Y, g_Z) before time loop (to get from g field by the user)
	double g_X = 0.; double g_Y = 0.; double g_Z = 0.;

	// An Intermediate Solid velocity (for velocity-verlet algorithm purposes)
	double US2_X = 0.; double US2_Y = 0.; double US2_Z = 0.;

	
	// ***************************  INITIALIZATION START (get US and g from user) ***************************/
	forAll(mesh.cells(),i)
	{
	US_X = Us[i].component(0) ;  g_X = g[i].component(0) ;
	US_Y = Us[i].component(1) ;  g_Y = g[i].component(1) ;
	US_Z = Us[i].component(2) ;  g_Z = g[i].component(2) ;
	}
	// ***************************  INITIALIZATION END ******************************************************/


    for (runTime++; !runTime.end(); runTime++)
    {
        Info<< "Time = " << runTime.timeName() << nl << endl;



#       include "CourantNo.H"



					///////////////////////////////////////////  Velocity-Verlet Algorithm //////////////////////////////////////////////////////////////////////////////
					/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					// //////////////////////////////////////////////  time step: n /////////////////////////////////////////////////////////////////////////////////////
					/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


					/****************************************************************************************************************************************************/
					/*********** We know at this stage all of U(Ux, Uy, Uz), US(US_X, US_Y, US_Z) & FS(FS_X, FS_Y, FS_Z) at time step: n  *******************************/
					/****************************************************************************************************************************************************/
			//////////////////////////////// To Activate only if moving body //////////////////////////////////////////////////////	
			/*
				if( count % COUNT == 0 )
				{
					// Update (xc, yc,zc) due to the (US_X, US_Y, US_Z) & (FS_X, FS_Y, FS_Z) at time step: n
					xc.value() = xc.value() + COUNT*US_X*runTime.deltaTValue() + 0.5*(FS_X/mass.value())*rhof.value()*runTime.deltaTValue()*runTime.deltaTValue() ;
					yc.value() = yc.value() + COUNT*US_Y*runTime.deltaTValue() + 0.5*(FS_Y/mass.value())*rhof.value()*runTime.deltaTValue()*runTime.deltaTValue() ;
					zc.value() = zc.value() + COUNT*US_Z*runTime.deltaTValue() + 0.5*(FS_Z/mass.value())*rhof.value()*runTime.deltaTValue()*runTime.deltaTValue() ;
				}
			*/

					// Intermediate solid velocity calculation (time step: n +(1/2) ) due to (US_X, US_Y, US_Z) & (FS_X, FS_Y, FS_Z) at time step: n
					US2_X = US_X + 0.5*(FS_X/mass.value())*rhof.value()*runTime.deltaTValue() ;
					US2_Y = US_Y + 0.5*(FS_Y/mass.value())*rhof.value()*runTime.deltaTValue() ;
					US2_Z = US_Z + 0.5*(FS_Z/mass.value())*rhof.value()*runTime.deltaTValue() ;
					/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



	/****************************************************************************************************************/
	//*************** Fill and update both etta and Us FILEDS in the domain before each time step *******************/
	/****************************************************************************************************************/
	forAll(mesh.cells(),i)
	{

	xdim=mesh.C()[i].component(0);
	ydim=mesh.C()[i].component(1);
	zdim=mesh.C()[i].component(2);
	if (       (  ((xdim.value()-xc.value())*(xdim.value()-xc.value()))
	             +((ydim.value()-yc.value())*(ydim.value()-yc.value()))
	             +((zdim.value()-zc.value())*(zdim.value()-zc.value()))  ) <= (Radius.value()*Radius.value())      ) 
	{
	etta[i]=(1.e0 - 1.e-12);
	oneminusetta[i]=(1.e-12);
	// Assign (US_X, US_Y, US_Z) to Us_FIELD
	Us[i].component(0) = US_X;
	Us[i].component(1) = US_Y;
	Us[i].component(2) = US_Z;
	}
	else
	{
	etta[i]=(1.e-12);
	oneminusetta[i]=(1.e0 - 1.e-12);
	// Assign (0, 0, 0) to Us_FIELD
	Us[i].component(0) = 0.;
	Us[i].component(1) = 0.;
	Us[i].component(2) = 0.;
	}

	} 
	/****************************************************************************************************************/



if (nbc==0)
{

	    rho=rhop*c+rhof*(1.0-c); // bouyancy effects added

            // Momentum predictor

            tmp<fvVectorMatrix> UEqn
            (
    
              fvm::ddt(U) +  oneminusetta*fvm::div(phi,U) + oneminusetta*(- fvm::laplacian(nu, U) -fvc::div(nu*dev2(fvc::grad(U)().T())) - fvc::div(SigmaL) ) // changed lastly (dev2) added
            );

            UEqn().relax();

            solve(UEqn() == -oneminusetta*fvc::grad(p) + oneminusetta*((rhof-rhop)/rho)*g*c + etta*((Us-U)/(runTime.deltaTValue()*dtunit)) ); // changed lastly


        // --- PISO loop
         while (piso.correct())
        {




            volScalarField rUA = 1.0/UEqn().A();
            U = rUA*UEqn().H();

            ff = etta*(  fvc::grad(p) + UEqn().A()*(Us-U)  ); 


            phi = fvc::interpolate(U) & mesh.Sf();
            adjustPhi(phi, U, p);

            // Store pressure for under-relaxation
            p.storePrevIter();

            // Non-orthogonal pressure corrector loop
            while (piso.correctNonOrthogonal())
            {
                fvScalarMatrix pEqn
                (
                    fvm::laplacian(rUA, p) == fvc::div(phi)
                );

                pEqn.setReference(pRefCell, pRefValue);
                pEqn.solve(mesh.solver(p.select(piso.finalInnerIter())));

                if (piso.finalNonOrthogonalIter())
                {
                    phi -= pEqn.flux();
                }
            }

#           include "continuityErrs.H"

            // Explicitly relax pressure for momentum corrector
            p.relax();

            // Momentum corrector
            //U -= rUA*fvc::grad(p);
	    U = U + rUA*(ff - fvc::grad(p));
            U.correctBoundaryConditions();
        }

}
nbc=1;


///////////////////////////////////////// Added Part /////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

Gr=fvc::grad(U);
E = 0.5*(Gr+(Gr.T()));
gammadot = sqrt( (2.0*(E && E)) );

nuS= pow( (1.0- (c/cm)), -2.0); // suspension normalized viscosity (no unit)
nuN= 1.07648 * pow((c/cm),2) * nuS; // normal stress normalized viscosity (no unit)

nu=nu0*nuS; // suspension viscosity (Pa.s)
nu2=nu0*nuN; // normal stress viscosity (Pa.s)

fc=(1.0-(c/cm))*pow((1.0-c),alpha-1.0); //Hyndrance function

rho=rhop*c+rhof*(1.0-c); // bouyancy effects added

volScalarField lambda2=0.812*(c/cm)+0.665; // lambda2 (2nd diagonal component in Tensor Q)
volScalarField lambda3=-0.008874*(c/cm)+0.547; // lambda3 (3rd diagonal component in Tensor Q)


Umax=max(max(U.component(0),U.component(1)),U.component(2));

EnoDim=E & II; // to make assymmetric and remove dimensions use EnoDim


myEigenVectors = eigenVectors(EnoDim); // it is the transformation matrix



myEigenValues=eigenValues(EnoDim);





///////////////////////////////////// lamda1 and lamda2 , and EigenVectors E1 and E2 calculation /////////////////////////////////////////////////////

// internal field getting delta
forAll(mesh.cells(),i)
{
//a=1;
bb[i] = EnoDim[i].component(tensor::XX) + EnoDim[i].component(tensor::YY);
cc[i] = ( EnoDim[i].component(tensor::XX) * EnoDim[i].component(tensor::YY) ) - pow(EnoDim[i].component(tensor::XY),2);
delta[i] = (bb[i] * bb[i]) - (4.0 * cc[i]);
}
// boundary field getting delta
// b1
forAll(mesh.boundaryMesh()[b1],i)
{
//a=1;
bb.boundaryField()[b1][i] = EnoDim.boundaryField()[b1][i].component(tensor::XX) + EnoDim.boundaryField()[b1][i].component(tensor::YY);
cc.boundaryField()[b1][i] = ( EnoDim.boundaryField()[b1][i].component(tensor::XX) * EnoDim.boundaryField()[b1][i].component(tensor::YY) ) - pow(EnoDim.boundaryField()[b1][i].component(tensor::XY),2);
delta.boundaryField()[b1][i] = (bb.boundaryField()[b1][i] * bb.boundaryField()[b1][i]) - (4.0 * cc.boundaryField()[b1][i]);
}
// b2
forAll(mesh.boundaryMesh()[b2],i)
{
//a=1;
bb.boundaryField()[b2][i] = EnoDim.boundaryField()[b2][i].component(tensor::XX) + EnoDim.boundaryField()[b2][i].component(tensor::YY);
cc.boundaryField()[b2][i] = ( EnoDim.boundaryField()[b2][i].component(tensor::XX) * EnoDim.boundaryField()[b2][i].component(tensor::YY) ) - pow(EnoDim.boundaryField()[b2][i].component(tensor::XY),2);
delta.boundaryField()[b2][i] = (bb.boundaryField()[b2][i] * bb.boundaryField()[b2][i]) - (4.0 * cc.boundaryField()[b2][i]);
}
// b3
forAll(mesh.boundaryMesh()[b3],i)
{
//a=1;
bb.boundaryField()[b3][i] = EnoDim.boundaryField()[b3][i].component(tensor::XX) + EnoDim.boundaryField()[b3][i].component(tensor::YY);
cc.boundaryField()[b3][i] = ( EnoDim.boundaryField()[b3][i].component(tensor::XX) * EnoDim.boundaryField()[b3][i].component(tensor::YY) ) - pow(EnoDim.boundaryField()[b3][i].component(tensor::XY),2);
delta.boundaryField()[b3][i] = (bb.boundaryField()[b3][i] * bb.boundaryField()[b3][i]) - (4.0 * cc.boundaryField()[b3][i]);
}// b4
forAll(mesh.boundaryMesh()[b4],i)
{
//a=1;
bb.boundaryField()[b4][i] = EnoDim.boundaryField()[b4][i].component(tensor::XX) + EnoDim.boundaryField()[b4][i].component(tensor::YY);
cc.boundaryField()[b4][i] = ( EnoDim.boundaryField()[b4][i].component(tensor::XX) * EnoDim.boundaryField()[b4][i].component(tensor::YY) ) - pow(EnoDim.boundaryField()[b4][i].component(tensor::XY),2);
delta.boundaryField()[b4][i] = (bb.boundaryField()[b4][i] * bb.boundaryField()[b4][i]) - (4.0 * cc.boundaryField()[b4][i]);
}

//calculating lamda1 and lamda2 for all Mesh
lamda1= ( bb + sqrt(delta) )/2;
lamda2= ( bb - sqrt(delta) )/2;




//calculating E1 and E2 EigenVectors corresponding to lamda1 and lamda2, respectively

// internal field Eigen vector E1 
forAll(mesh.cells(),i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim[i].component(tensor::XX) - lamda1[i];
if (mag(EnoDim[i])>1e-5)
{
if ( mag(dd) > 1e-4 )
{
ey = 1.0; ex = (- EnoDim[i].component(tensor::XY) * ey)/( dd );
}
else
{
//dd = EnoDim[i].component(tensor::YY) - lamda1[i];
ex = 1.0; ey = (- EnoDim[i].component(tensor::XY) * ex)/( EnoDim[i].component(tensor::YY) - lamda1[i]);
}

float ez=0.0;
E1[i].component(vector::X) = ex;
E1[i].component(vector::Y) = ey;
E1[i].component(vector::Z) = ez;
}

if (mag(EnoDim[i])<1e-5)
{
E1[i].component(vector::X) = 1.;
E1[i].component(vector::Y) = 0.;
E1[i].component(vector::Z) = 0.;
}
}


// Boundary field Eigen vector E1 
//b1
forAll(mesh.boundaryMesh()[b1],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b1][i].component(tensor::XX) - lamda1.boundaryField()[b1][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b1][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b1][i].component(tensor::YY) - lamda1.boundaryField()[b1][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b1][i].component(tensor::XY) * ex)/( dd );  
}

float ez=0.0;
E1.boundaryField()[b1][i].component(vector::X) = ex;
E1.boundaryField()[b1][i].component(vector::Y) = ey;
E1.boundaryField()[b1][i].component(vector::Z) = ez;
}
//b2
forAll(mesh.boundaryMesh()[b2],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b2][i].component(tensor::XX) - lamda1.boundaryField()[b2][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b2][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b2][i].component(tensor::YY) - lamda1.boundaryField()[b2][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b2][i].component(tensor::XY) * ex)/( dd ); 
}

float ez=0.0;
E1.boundaryField()[b2][i].component(vector::X) = ex;
E1.boundaryField()[b2][i].component(vector::Y) = ey;
E1.boundaryField()[b2][i].component(vector::Z) = ez;
}
//b3
forAll(mesh.boundaryMesh()[b3],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b3][i].component(tensor::XX) - lamda1.boundaryField()[b3][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b3][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b3][i].component(tensor::YY) - lamda1.boundaryField()[b3][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b3][i].component(tensor::XY) * ex)/( dd ); 
}

float ez=0.0;
E1.boundaryField()[b3][i].component(vector::X) = ex;
E1.boundaryField()[b3][i].component(vector::Y) = ey;
E1.boundaryField()[b3][i].component(vector::Z) = ez;
}//b4
forAll(mesh.boundaryMesh()[b4],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b4][i].component(tensor::XX) - lamda1.boundaryField()[b4][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b4][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b4][i].component(tensor::YY) - lamda1.boundaryField()[b4][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b4][i].component(tensor::XY) * ex)/( dd ); 
}

float ez=0.0;
E1.boundaryField()[b4][i].component(vector::X) = ex;
E1.boundaryField()[b4][i].component(vector::Y) = ey;
E1.boundaryField()[b4][i].component(vector::Z) = ez;
}


// internal field Eigen vector E2 
forAll(mesh.cells(),i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim[i].component(tensor::XX) - lamda2[i];
if (mag(EnoDim[i])>1e-5)
{
if ( mag(dd) > 1e-4 )
{
ey = 1.0; ex = (- EnoDim[i].component(tensor::XY) * ey)/( dd );
}
else
{
//dd = EnoDim[i].component(tensor::YY) - lamda1[i];
ex = 1.0; ey = (- EnoDim[i].component(tensor::XY) * ex)/( EnoDim[i].component(tensor::YY) - lamda2[i]);
}

float ez=0.0;
E2[i].component(vector::X) = ex;
E2[i].component(vector::Y) = ey;
E2[i].component(vector::Z) = ez;
}

if (mag(EnoDim[i])<1e-5)
{
E2[i].component(vector::X) = 1.;
E2[i].component(vector::Y) = 0.;
E2[i].component(vector::Z) = 0.;
}
}

// Boundary field Eigen vector E2 
//b1
forAll(mesh.boundaryMesh()[b1],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b1][i].component(tensor::XX) - lamda2.boundaryField()[b1][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b1][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b1][i].component(tensor::YY) - lamda2.boundaryField()[b1][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b1][i].component(tensor::XY) * ex)/( dd );  
}

float ez=0.0;
E2.boundaryField()[b1][i].component(vector::X) = ex;
E2.boundaryField()[b1][i].component(vector::Y) = ey;
E2.boundaryField()[b1][i].component(vector::Z) = ez;
}
//b2
forAll(mesh.boundaryMesh()[b2],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b2][i].component(tensor::XX) - lamda2.boundaryField()[b2][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b2][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b2][i].component(tensor::YY) - lamda2.boundaryField()[b2][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b2][i].component(tensor::XY) * ex)/( dd ); 
}

float ez=0.0;
E2.boundaryField()[b2][i].component(vector::X) = ex;
E2.boundaryField()[b2][i].component(vector::Y) = ey;
E2.boundaryField()[b2][i].component(vector::Z) = ez;
}
//b3
forAll(mesh.boundaryMesh()[b3],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b3][i].component(tensor::XX) - lamda2.boundaryField()[b3][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b3][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b3][i].component(tensor::YY) - lamda2.boundaryField()[b3][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b3][i].component(tensor::XY) * ex)/( dd ); 
}

float ez=0.0;
E2.boundaryField()[b3][i].component(vector::X) = ex;
E2.boundaryField()[b3][i].component(vector::Y) = ey;
E2.boundaryField()[b3][i].component(vector::Z) = ez;
}
//b4
forAll(mesh.boundaryMesh()[b4],i)
{
float ey = 1.0; float ex = 1.0;
float dd = EnoDim.boundaryField()[b4][i].component(tensor::XX) - lamda2.boundaryField()[b4][i];
if ( mag(dd) > 1e-10 )
{
ey = 1.0; ex = (- EnoDim.boundaryField()[b4][i].component(tensor::XY) * ey)/( dd ); 
}
else
{
dd=EnoDim.boundaryField()[b4][i].component(tensor::YY) - lamda2.boundaryField()[b4][i];
ex = 1.0; ey = (- EnoDim.boundaryField()[b4][i].component(tensor::XY) * ex)/( dd ); 
}

float ez=0.0;
E2.boundaryField()[b4][i].component(vector::X) = ex;
E2.boundaryField()[b4][i].component(vector::Y) = ey;
E2.boundaryField()[b4][i].component(vector::Z) = ez;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


E1=E1/mag(E1);
E2=E2/mag(E2);

// orientation de et et ec

et=E1;
ec=E2;

// on veut et suivant la vitesse
volVectorField UnoDim=IIIII*U;
check = (et & UnoDim);

forAll(mesh.cells(),i)
{
if  (check[i] < 0) 
{
et[i].component(vector::X)=-et[i].component(vector::X);
et[i].component(vector::Y)=-et[i].component(vector::Y);
et[i].component(vector::Z)=-et[i].component(vector::Z);
}
}

//b1
forAll(mesh.boundaryMesh()[b1],i)
{
if  (check.boundaryField()[b1][i] < 0) 
{
et.boundaryField()[b1][i].component(vector::X)=-et.boundaryField()[b1][i].component(vector::X);
et.boundaryField()[b1][i].component(vector::Y)=-et.boundaryField()[b1][i].component(vector::Y);
et.boundaryField()[b1][i].component(vector::Z)=-et.boundaryField()[b1][i].component(vector::Z);
}
}
//b2
forAll(mesh.boundaryMesh()[b2],i)
{
if  (check.boundaryField()[b2][i] < 0) 
{
et.boundaryField()[b2][i].component(vector::X)=-et.boundaryField()[b2][i].component(vector::X);
et.boundaryField()[b2][i].component(vector::Y)=-et.boundaryField()[b2][i].component(vector::Y);
et.boundaryField()[b2][i].component(vector::Z)=-et.boundaryField()[b2][i].component(vector::Z);
}
}
//b3
forAll(mesh.boundaryMesh()[b3],i)
{
if  (check.boundaryField()[b3][i] < 0) 
{
et.boundaryField()[b3][i].component(vector::X)=-et.boundaryField()[b3][i].component(vector::X);
et.boundaryField()[b3][i].component(vector::Y)=-et.boundaryField()[b3][i].component(vector::Y);
et.boundaryField()[b3][i].component(vector::Z)=-et.boundaryField()[b3][i].component(vector::Z);
}
}
//b4
forAll(mesh.boundaryMesh()[b4],i)
{
if  (check.boundaryField()[b4][i] < 0) 
{
et.boundaryField()[b4][i].component(vector::X)=-et.boundaryField()[b4][i].component(vector::X);
et.boundaryField()[b4][i].component(vector::Y)=-et.boundaryField()[b4][i].component(vector::Y);
et.boundaryField()[b4][i].component(vector::Z)=-et.boundaryField()[b4][i].component(vector::Z);
}
}
// on veur et ^ ec suivant -vorticity

w=fvc::curl(U);

wnodim = w * IIII;

check = (et ^ ec) & wnodim;

forAll(mesh.cells(),i)
{
if  (check[i] < 0) 
{
ec[i].component(vector::X)=-ec[i].component(vector::X);
ec[i].component(vector::Y)=-ec[i].component(vector::Y);
ec[i].component(vector::Z)=-ec[i].component(vector::Z);
}
}

//b1
forAll(mesh.boundaryMesh()[b1],i)
{
if  (check.boundaryField()[b1][i] < 0) 
{
ec.boundaryField()[b1][i].component(vector::X)=-ec.boundaryField()[b1][i].component(vector::X);
ec.boundaryField()[b1][i].component(vector::Y)=-ec.boundaryField()[b1][i].component(vector::Y);
ec.boundaryField()[b1][i].component(vector::Z)=-ec.boundaryField()[b1][i].component(vector::Z);
}
}
//b2
forAll(mesh.boundaryMesh()[b2],i)
{
if  (check.boundaryField()[b2][i] < 0) 
{
ec.boundaryField()[b2][i].component(vector::X)=-ec.boundaryField()[b2][i].component(vector::X);
ec.boundaryField()[b2][i].component(vector::Y)=-ec.boundaryField()[b2][i].component(vector::Y);
ec.boundaryField()[b2][i].component(vector::Z)=-ec.boundaryField()[b2][i].component(vector::Z);
}
}
//b3
forAll(mesh.boundaryMesh()[b3],i)
{
if  (check.boundaryField()[b3][i] < 0) 
{
ec.boundaryField()[b3][i].component(vector::X)=-ec.boundaryField()[b3][i].component(vector::X);
ec.boundaryField()[b3][i].component(vector::Y)=-ec.boundaryField()[b3][i].component(vector::Y);
ec.boundaryField()[b3][i].component(vector::Z)=-ec.boundaryField()[b3][i].component(vector::Z);
}
}
//b4
forAll(mesh.boundaryMesh()[b4],i)
{
if  (check.boundaryField()[b4][i] < 0) 
{
ec.boundaryField()[b4][i].component(vector::X)=-ec.boundaryField()[b4][i].component(vector::X);
ec.boundaryField()[b4][i].component(vector::Y)=-ec.boundaryField()[b4][i].component(vector::Y);
ec.boundaryField()[b4][i].component(vector::Z)=-ec.boundaryField()[b4][i].component(vector::Z);
}
}
check = (et ^ ec) & wnodim;

//Transformation Matrix myP = Eigen vectors et, ec, and (0 0 1)
// Internal field myP
forAll(mesh.cells(),i)
{
myP[i].component(tensor::XX) = et[i].component(vector::X);
myP[i].component(tensor::YX) = et[i].component(vector::Y);
myP[i].component(tensor::ZX) = et[i].component(vector::Z);

myP[i].component(tensor::XY) = ec[i].component(vector::X);
myP[i].component(tensor::YY) = ec[i].component(vector::Y);
myP[i].component(tensor::ZY) = ec[i].component(vector::Z);

myP[i].component(tensor::XZ) = 0.;
myP[i].component(tensor::YZ) = 0.;
myP[i].component(tensor::ZZ) = 1.;
}
// Boundary field myP
//b1
forAll(mesh.boundaryMesh()[b1],i)
{
myP.boundaryField()[b1][i].component(tensor::XX) = et.boundaryField()[b1][i].component(vector::X);
myP.boundaryField()[b1][i].component(tensor::YX) = et.boundaryField()[b1][i].component(vector::Y);
myP.boundaryField()[b1][i].component(tensor::ZX) = et.boundaryField()[b1][i].component(vector::Z);

myP.boundaryField()[b1][i].component(tensor::XY) = ec.boundaryField()[b1][i].component(vector::X);
myP.boundaryField()[b1][i].component(tensor::YY) = ec.boundaryField()[b1][i].component(vector::Y);
myP.boundaryField()[b1][i].component(tensor::ZY) = ec.boundaryField()[b1][i].component(vector::Z);

myP.boundaryField()[b1][i].component(tensor::XZ) = 0.;
myP.boundaryField()[b1][i].component(tensor::YZ) = 0.;
myP.boundaryField()[b1][i].component(tensor::ZZ) = 1.;
}
//b2
forAll(mesh.boundaryMesh()[b2],i)
{
myP.boundaryField()[b2][i].component(tensor::XX) = et.boundaryField()[b2][i].component(vector::X);
myP.boundaryField()[b2][i].component(tensor::YX) = et.boundaryField()[b2][i].component(vector::Y);
myP.boundaryField()[b2][i].component(tensor::ZX) = et.boundaryField()[b2][i].component(vector::Z);

myP.boundaryField()[b2][i].component(tensor::XY) = ec.boundaryField()[b2][i].component(vector::X);
myP.boundaryField()[b2][i].component(tensor::YY) = ec.boundaryField()[b2][i].component(vector::Y);
myP.boundaryField()[b2][i].component(tensor::ZY) = ec.boundaryField()[b2][i].component(vector::Z);

myP.boundaryField()[b2][i].component(tensor::XZ) = 0.;
myP.boundaryField()[b2][i].component(tensor::YZ) = 0.;
myP.boundaryField()[b2][i].component(tensor::ZZ) = 1.;
}
//b3
forAll(mesh.boundaryMesh()[b3],i)
{
myP.boundaryField()[b3][i].component(tensor::XX) = et.boundaryField()[b3][i].component(vector::X);
myP.boundaryField()[b3][i].component(tensor::YX) = et.boundaryField()[b3][i].component(vector::Y);
myP.boundaryField()[b3][i].component(tensor::ZX) = et.boundaryField()[b3][i].component(vector::Z);

myP.boundaryField()[b3][i].component(tensor::XY) = ec.boundaryField()[b3][i].component(vector::X);
myP.boundaryField()[b3][i].component(tensor::YY) = ec.boundaryField()[b3][i].component(vector::Y);
myP.boundaryField()[b3][i].component(tensor::ZY) = ec.boundaryField()[b3][i].component(vector::Z);

myP.boundaryField()[b3][i].component(tensor::XZ) = 0.;
myP.boundaryField()[b3][i].component(tensor::YZ) = 0.;
myP.boundaryField()[b3][i].component(tensor::ZZ) = 1.;
}
//b4
forAll(mesh.boundaryMesh()[b4],i)
{
myP.boundaryField()[b4][i].component(tensor::XX) = et.boundaryField()[b4][i].component(vector::X);
myP.boundaryField()[b4][i].component(tensor::YX) = et.boundaryField()[b4][i].component(vector::Y);
myP.boundaryField()[b4][i].component(tensor::ZX) = et.boundaryField()[b4][i].component(vector::Z);

myP.boundaryField()[b4][i].component(tensor::XY) = ec.boundaryField()[b4][i].component(vector::X);
myP.boundaryField()[b4][i].component(tensor::YY) = ec.boundaryField()[b4][i].component(vector::Y);
myP.boundaryField()[b4][i].component(tensor::ZY) = ec.boundaryField()[b4][i].component(vector::Z);

myP.boundaryField()[b4][i].component(tensor::XZ) = 0.;
myP.boundaryField()[b4][i].component(tensor::YZ) = 0.;
myP.boundaryField()[b4][i].component(tensor::ZZ) = 1.;
}



total=(fvc::ddt(et)+fvc::div(phi, et));
Wb=et ^ total;


Wrel=(w/2.0)-Wb;


rhok=(2.0*mag(Wrel)/max(WSmall, (gammadot/2)+mag(Wrel)));





//Tensor Q

// assign Q on internal Field

/////////////////////////////////////////////////////////
forAll(mesh.cells(),j)
{
Q[j].component(tensor::XX) = 1;
Q[j].component(tensor::XY) = 0;
Q[j].component(tensor::XZ) = 0;
Q[j].component(tensor::YX) = 0;
Q[j].component(tensor::YY) = lambda2[j]; 
Q[j].component(tensor::YZ) = 0;
Q[j].component(tensor::ZX) = 0;
Q[j].component(tensor::ZY) = 0;
Q[j].component(tensor::ZZ) = lambda3[j];
}
//////////////////////////////////////////////////////////

// assign Q on boundaries

// b1
forAll(mesh.boundaryMesh()[b1],i)
{
Q.boundaryField()[b1][i].component(tensor::XX) = 1;
Q.boundaryField()[b1][i].component(tensor::XY) = 0;
Q.boundaryField()[b1][i].component(tensor::XZ) = 0;
Q.boundaryField()[b1][i].component(tensor::YX) = 0;
Q.boundaryField()[b1][i].component(tensor::YY) = lambda2.boundaryField()[b1][i];
Q.boundaryField()[b1][i].component(tensor::YZ) = 0;
Q.boundaryField()[b1][i].component(tensor::ZX) = 0;
Q.boundaryField()[b1][i].component(tensor::ZY) = 0;
Q.boundaryField()[b1][i].component(tensor::ZZ) = lambda3.boundaryField()[b1][i];
} 
// b2
forAll(mesh.boundaryMesh()[b2],i)
{
Q.boundaryField()[b2][i].component(tensor::XX) = 1;
Q.boundaryField()[b2][i].component(tensor::XY) = 0;
Q.boundaryField()[b2][i].component(tensor::XZ) = 0;
Q.boundaryField()[b2][i].component(tensor::YX) = 0;
Q.boundaryField()[b2][i].component(tensor::YY) = lambda2.boundaryField()[b2][i];
Q.boundaryField()[b2][i].component(tensor::YZ) = 0;
Q.boundaryField()[b2][i].component(tensor::ZX) = 0;
Q.boundaryField()[b2][i].component(tensor::ZY) = 0;
Q.boundaryField()[b2][i].component(tensor::ZZ) = lambda3.boundaryField()[b2][i];
} 
// b3
forAll(mesh.boundaryMesh()[b3],i)
{
Q.boundaryField()[b3][i].component(tensor::XX) = 1;
Q.boundaryField()[b3][i].component(tensor::XY) = 0;
Q.boundaryField()[b3][i].component(tensor::XZ) = 0;
Q.boundaryField()[b3][i].component(tensor::YX) = 0;
Q.boundaryField()[b3][i].component(tensor::YY) = lambda2.boundaryField()[b3][i];
Q.boundaryField()[b3][i].component(tensor::YZ) = 0;
Q.boundaryField()[b3][i].component(tensor::ZX) = 0;
Q.boundaryField()[b3][i].component(tensor::ZY) = 0;
Q.boundaryField()[b3][i].component(tensor::ZZ) = lambda3.boundaryField()[b3][i];
} 
// b4
forAll(mesh.boundaryMesh()[b4],i)
{
Q.boundaryField()[b4][i].component(tensor::XX) = 1;
Q.boundaryField()[b4][i].component(tensor::XY) = 0;
Q.boundaryField()[b4][i].component(tensor::XZ) = 0;
Q.boundaryField()[b4][i].component(tensor::YX) = 0;
Q.boundaryField()[b4][i].component(tensor::YY) = lambda2.boundaryField()[b4][i];
Q.boundaryField()[b4][i].component(tensor::YZ) = 0;
Q.boundaryField()[b4][i].component(tensor::ZX) = 0;
Q.boundaryField()[b4][i].component(tensor::ZY) = 0;
Q.boundaryField()[b4][i].component(tensor::ZZ) = lambda3.boundaryField()[b4][i];
} 

////////////////////////////// SigmaL ////////////////////////////////////////////////////////////////////////////////
SigmaLprim = - nu2 *( gammadot+ (epsilon*Umax) ) * Q;//epsilon=a/(H/2)² in case of channel flows and zero elsewhere
SigmaR = 2.0 * (nu - nu0)  * E;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



// SigmaL on internal Field
/////////////////////////////////////////////////////////
forAll(mesh.cells(),i)
{
float Bt=1.0;
float Bc=1.0;

SigmaL[i].component(tensor::XX) = ((SigmaLprim[i].component(tensor::XX)+SigmaLprim[i].component(tensor::YY))/2.0)*Bt; 
SigmaL[i].component(tensor::XY) = ((SigmaLprim[i].component(tensor::XX)-SigmaLprim[i].component(tensor::YY))/2.0)*Bt; 
SigmaL[i].component(tensor::XZ) = 0;
SigmaL[i].component(tensor::YX) = ((SigmaLprim[i].component(tensor::XX)-SigmaLprim[i].component(tensor::YY))/2.0)*Bt; 
SigmaL[i].component(tensor::YY) = ((SigmaLprim[i].component(tensor::XX)+SigmaLprim[i].component(tensor::YY))/2.0)*Bc; 
SigmaL[i].component(tensor::YZ) = 0;
SigmaL[i].component(tensor::ZX) = 0; 
SigmaL[i].component(tensor::ZY) = 0; 
SigmaL[i].component(tensor::ZZ) = SigmaLprim[i].component(tensor::ZZ);


if (rhok[i]<1)
{
Bt=1.0*rhok[i];
Bc=2.0-1.0*rhok[i];

SigmaL[i].component(tensor::XX) = ((SigmaLprim[i].component(tensor::XX)+SigmaLprim[i].component(tensor::YY))/2.0)*Bt; 
SigmaL[i].component(tensor::XY) = ((SigmaLprim[i].component(tensor::XX)-SigmaLprim[i].component(tensor::YY))/2.0)*Bt; 
SigmaL[i].component(tensor::XZ) = 0;
SigmaL[i].component(tensor::YX) = ((SigmaLprim[i].component(tensor::XX)-SigmaLprim[i].component(tensor::YY))/2.0)*Bt; 
SigmaL[i].component(tensor::YY) = ((SigmaLprim[i].component(tensor::XX)+SigmaLprim[i].component(tensor::YY))/2.0)*Bc; 
SigmaL[i].component(tensor::YZ) = 0;
SigmaL[i].component(tensor::ZX) = 0; 
SigmaL[i].component(tensor::ZY) = 0; 
SigmaL[i].component(tensor::ZZ) = SigmaLprim[i].component(tensor::ZZ);
}


}
//////////////////////////////////////////////////////////


// SigmaL on boundaries



// b1
forAll(mesh.boundaryMesh()[b1],i)
{
float Bt=1.0;
float Bc=1.0;

SigmaL.boundaryField()[b1][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)+SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b1][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)-SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b1][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b1][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)-SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b1][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)+SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b1][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b1][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b1][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b1][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b1][i].component(tensor::ZZ);

if (rhok[i]<1)
{
Bt=1.0*rhok.boundaryField()[b1][i];
Bc=2.0-1.0*rhok.boundaryField()[b1][i];

SigmaL.boundaryField()[b1][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)+SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b1][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)-SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b1][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b1][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)-SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b1][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b1][i].component(tensor::XX)+SigmaLprim.boundaryField()[b1][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b1][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b1][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b1][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b1][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b1][i].component(tensor::ZZ);
}

}

// b2
forAll(mesh.boundaryMesh()[b2],i)
{
float Bt=1.0;
float Bc=1.0;

SigmaL.boundaryField()[b2][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)+SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b2][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)-SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b2][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b2][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)-SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b2][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)+SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b2][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b2][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b2][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b2][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b2][i].component(tensor::ZZ);

if (rhok[i]<1)
{
Bt=1.0*rhok.boundaryField()[b2][i];
Bc=2.0-1.0*rhok.boundaryField()[b2][i];

SigmaL.boundaryField()[b2][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)+SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b2][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)-SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b2][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b2][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)-SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b2][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b2][i].component(tensor::XX)+SigmaLprim.boundaryField()[b2][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b2][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b2][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b2][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b2][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b2][i].component(tensor::ZZ);
}

}

// b3
forAll(mesh.boundaryMesh()[b3],i)
{
float Bt=1.0;
float Bc=1.0;

SigmaL.boundaryField()[b3][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)+SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b3][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)-SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b3][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b3][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)-SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b3][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)+SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b3][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b3][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b3][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b3][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b3][i].component(tensor::ZZ);

if (rhok[i]<1)
{
Bt=1.0*rhok.boundaryField()[b3][i];
Bc=2.0-1.0*rhok.boundaryField()[b3][i];

SigmaL.boundaryField()[b3][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)+SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b3][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)-SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b3][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b3][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)-SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b3][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b3][i].component(tensor::XX)+SigmaLprim.boundaryField()[b3][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b3][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b3][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b3][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b3][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b3][i].component(tensor::ZZ);
}

}

// b4
forAll(mesh.boundaryMesh()[b4],i)
{
float Bt=1.0;
float Bc=1.0;

SigmaL.boundaryField()[b4][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)+SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b4][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)-SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b4][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b4][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)-SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b4][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)+SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b4][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b4][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b4][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b4][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b4][i].component(tensor::ZZ);

if (rhok[i]<1)
{
Bt=1.0*rhok.boundaryField()[b4][i];
Bc=2.0-1.0*rhok.boundaryField()[b4][i];

SigmaL.boundaryField()[b4][i].component(tensor::XX) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)+SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bt; 
SigmaL.boundaryField()[b4][i].component(tensor::XY) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)-SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b4][i].component(tensor::XZ) = 0;
SigmaL.boundaryField()[b4][i].component(tensor::YX) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)-SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bt;
SigmaL.boundaryField()[b4][i].component(tensor::YY) = ((SigmaLprim.boundaryField()[b4][i].component(tensor::XX)+SigmaLprim.boundaryField()[b4][i].component(tensor::YY))/2.0)*Bc; 
SigmaL.boundaryField()[b4][i].component(tensor::YZ) = 0;
SigmaL.boundaryField()[b4][i].component(tensor::ZX) = 0; 
SigmaL.boundaryField()[b4][i].component(tensor::ZY) = 0; 
SigmaL.boundaryField()[b4][i].component(tensor::ZZ) = SigmaLprim.boundaryField()[b4][i].component(tensor::ZZ);
}

}





//*********************************************************************************//

//change SigmaP to the original basis

SigmaL = ((myP) & SigmaL) & (myP.T()) ;

SigmaR = 2.0 * (nu - nu0)  * E;

SigmaP = SigmaL + SigmaR;

SigmaF = (-p*I) + (2.0*nu0*E);


Sigma = SigmaP + SigmaF; 

J = ((2.0*sqr(a))/(9.0*nu0))*(  fc*fvc::div(SigmaP) + fc*((rhop-rhof)/rhof)*g*c  );
/////////////////////////////////////////////////////////////////////////////

//*********************************************************************************************//       



	    rho=rhop*c+rhof*(1.0-c); // bouyancy effects added

            // Momentum predictor

            tmp<fvVectorMatrix> UEqn
            (
    
             fvm::ddt(U) +  oneminusetta*fvm::div(phi,U) + oneminusetta*(- fvm::laplacian(nu, U) -fvc::div(nu*dev2(fvc::grad(U)().T())) - fvc::div(SigmaL) ) // changed lastly (dev2) added
            );

            UEqn().relax();

            solve(UEqn() == -oneminusetta*fvc::grad(p) + oneminusetta*((rhof-rhop)/rho)*g*c + etta*((Us-U)/(runTime.deltaTValue()*dtunit)) ); // changed lastly

        // --- PISO loop
        while (piso.correct())
        {


            volScalarField rUA = 1.0/UEqn().A();
            U = rUA*UEqn().H();

            ff = etta*(  fvc::grad(p) + UEqn().A()*(Us-U)  ); 


            phi = fvc::interpolate(U) & mesh.Sf();
            adjustPhi(phi, U, p);

            // Store pressure for under-relaxation
            p.storePrevIter();

            // Non-orthogonal pressure corrector loop
            while (piso.correctNonOrthogonal())
            {
                fvScalarMatrix pEqn
                (
                    fvm::laplacian(rUA, p) == fvc::div(phi)
                );

                pEqn.setReference(pRefCell, pRefValue);
                pEqn.solve(mesh.solver(p.select(piso.finalInnerIter())));

                if (piso.finalNonOrthogonalIter())
                {
                    phi -= pEqn.flux();
                }
            }

#           include "continuityErrs.H"

            // Explicitly relax pressure for momentum corrector
            p.relax();

            // Momentum corrector
            //U -= rUA*fvc::grad(p);
	    U = U + rUA*(ff - fvc::grad(p));
            U.correctBoundaryConditions();
        }



///////////////  c Eqn :  /////////////////////////////////////////////////////////


	fvScalarMatrix cEqn
        (
            
          fvm::ddt(c) + oneminusetta*fvc::div(phi,c)
       
        );
	    

		solve (  cEqn == - oneminusetta*fvc::div(J) + etta*((c-cm)/(runTime.deltaTValue()*dtunit))  );

	forAll(mesh.cells(),i)
	{
	if (c[i]<=0.0000001)
	{
	c[i]=0.0000001;
	}
	}
	//b1
	forAll(mesh.boundaryMesh()[b1],i)
	{
	if  (c.boundaryField()[b1][i] <= 0.0000001) 
	{
	c.boundaryField()[b1][i]=0.0000001;
	}
	}
	//b2
	forAll(mesh.boundaryMesh()[b2],i)
	{
	if  (c.boundaryField()[b2][i] <= 0.0000001) 
	{
	c.boundaryField()[b2][i]=0.0000001;
	}
	}
	//b3
	forAll(mesh.boundaryMesh()[b3],i)
	{
	if  (c.boundaryField()[b3][i] <= 0.0000001) 
	{
	c.boundaryField()[b3][i]=0.0000001;
	}
	}
	//b4
	forAll(mesh.boundaryMesh()[b4],i)
	{
	if  (c.boundaryField()[b4][i] <= 0.0000001) 
	{
	c.boundaryField()[b4][i]=0.0000001;
	}
	}


	forAll(mesh.cells(),i)
	{
	if (c[i]>=cm.value()-0.03)
	{
	c[i]=cm.value()-0.03;
	}
	}
	//b1
	forAll(mesh.boundaryMesh()[b1],i)
	{
	if  (c.boundaryField()[b1][i] >= cm.value()-0.03) 
	{
	c.boundaryField()[b1][i]=cm.value()-0.03;
	}
	}
	//b2
	forAll(mesh.boundaryMesh()[b2],i)
	{
	if  (c.boundaryField()[b2][i] >= cm.value()-0.03) 
	{
	c.boundaryField()[b2][i]=cm.value()-0.03;
	}
	}
	//b3
	forAll(mesh.boundaryMesh()[b3],i)
	{
	if  (c.boundaryField()[b3][i] >= cm.value()-0.03) 
	{
	c.boundaryField()[b3][i]=cm.value()-0.03;
	}
	}
	//b4
	forAll(mesh.boundaryMesh()[b4],i)
	{
	if  (c.boundaryField()[b4][i] >= cm.value()-0.03) 
	{
	c.boundaryField()[b4][i]=cm.value()-0.03;
	}
	}	    		    	 
 
///////////////////////////////////////////////////////////////////////////////////

        runTime.write();

	count = count + 1;

        Info<< "ExecutionTime = "
            << runTime.elapsedCpuTime()
            << " s\n\n" << endl;

	// get ff at time step : n+1 due to U field at time step : n, and assign it to the FSI_Force  (forcing term at time step : n)
	ff = etta*(  fvc::div(phi, U) -fvc::laplacian(nu, U) -fvc::div(nu*dev2(fvc::grad(U)().T())) - fvc::div(SigmaL) + fvc::grad(p) +((Us-U)/(runTime.deltaTValue()*dtunit))  ) ;
	FSI_force = ff ;

	/****************************************************************************************************************************************************/
	/********** Now Calculation of Fdrag & Flift forces from  the FSI_force TERM and assign result to (FS2_X, FS2_Y, FS2_Z) at time step = n+1  *********/
	/****************************************************************************************************************************************************/
	
	double sumx = 0.; double sumy = 0. ; double sumz = 0. ;
	double FS2_X = 0.; double FS2_Y = 0. ; double FS2_Z = 0. ;

	const faceList & fl = mesh.faces();
	const pointField & pp = mesh.points();

	// Compute the Integral of FSI_Force TERM => sum on all cells  [ in order to get a force vector (FS2_X, FS2_Y, FS2_Z) ]
	forAll(mesh.C(), i)
	{

	    const cell & cc = mesh.cells()[i];
	    labelList pLabels(cc.labels(fl));
	    pointField pLocal(pLabels.size(), vector::zero);

	    forAll (pLabels, pointi)
            pLocal[pointi] = pp[pLabels[pointi]];

	    scalar xDim = Foam::max(pLocal & vector(1,0,0)) - Foam::min(pLocal & vector(1,0,0));
	    scalar yDim = Foam::max(pLocal & vector(0,1,0)) - Foam::min(pLocal & vector(0,1,0));
	    // if Domain is 2D	    
	       scalar zDim = 1. ;
	    // if Domain is 3D
	    // scalar zDim = Foam::max(pLocal & vector(0,0,1)) - Foam::min(pLocal & vector(0,0,1));
	
	sumx = sumx +    (xDim*yDim*zDim) * -FSI_force[i].component(0);
	sumy = sumy +    (xDim*yDim*zDim) * -FSI_force[i].component(1);
	sumz = sumz +    (xDim*yDim*zDim) * -FSI_force[i].component(2);

	} 

	// Assign result of the Integral to the force vector (FS2_X, FS2_Y, FS2_Z) [ for time step : n+1 ]
	FS2_X = sumx ; FS2_Y = sumy ; FS2_Z = sumz ;
	/***************************************************************************************************************************************************/




				//////////////////////////////// To Activate only if moving body //////////////////////////////////////////////////////
				//*******************************************************************************************************************//
				//****** Velocity-Verlet Algorithm continued ( update of solid velocity (US_X, US_Y, US_Z) at time step n+1 ) *******//
				//*******************************************************************************************************************//
/*				if( count % COUNT == 0 )
				{
					US_X = US2_X + 0.5*(FS2_X/mass.value())*rhof.value()*runTime.deltaTValue() + g_X*(1.-(rhof.value()/rhop.value()))*runTime.deltaTValue() ; 
					US_Y = US2_Y + 0.5*(FS2_Y/mass.value())*rhof.value()*runTime.deltaTValue() + g_Y*(1.-(rhof.value()/rhop.value()))*runTime.deltaTValue() ; 
					US_Z = US2_Z + 0.5*(FS2_Z/mass.value())*rhof.value()*runTime.deltaTValue() + g_Z*(1.-(rhof.value()/rhop.value()))*runTime.deltaTValue() ; 
				}
*/				////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		




	// Saving FSI and Us in File
	out << runTime.value() << " " << FS2_X << " " << FS2_Y << " " << FS2_Z << " " << US_X << " " << US_Y << " " << US_Z << std::endl;



				////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
				//Assign (FS2_X, FS2_Y, FS2_Z) into (FS_X, FS_Y, FS_Z) for n+1 <- n interchanging time steps
				FS_X = FS2_X ; FS_Y = FS2_Y ; FS_Z = FS2_Z ;
				////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    }

    Info<< "End\n" << endl;

    out.close();

    return(0);
}


// ************************************************************************* //
